# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''连接远程WebDriverAgent服务
'''

from __future__ import absolute_import, print_function

import socket
import string
import base64
import json

try:
    import http.client as httplib
    from urllib import request as url_request
    from urllib import parse
except ImportError: # above is available in py3+, below is py2.7
    import httplib as httplib
    import urllib2 as url_request
    import urlparse as parse

from testbase.conf import settings
from qt4i.driver.tools import logger as logging
from qt4i.driver.xctest.webdriverclient.command import Command
from qt4i.driver.xctest.webdriverclient.errorhandler import ErrorCode
from qt4i.driver.xctest.webdriverclient.exceptions import XCTestAgentTimeoutException


class Request(url_request.Request):
    """Extends the url_request.Request to support all HTTP request types.
    """

    def __init__(self, url, data=None, method=None):
        """Initialise a new HTTP request.

        :param url: String for the URL to send the request to.
        :param data: Data to send with the request.
        """
        if method is None:
            method = data is not None and 'POST' or 'GET'
        elif method != 'POST' and method != 'PUT':
            data = None
        self._method = method
        url_request.Request.__init__(self, url, data=data)

    def get_method(self):
        """Returns the HTTP method used by this request.
        """
        return self._method


class Response(object):
    """Represents an HTTP response.
    """

    def __init__(self, fp, code, headers, url):
        """Initialise a new Response.

        :param fp: The response body file object.
        :param code: The HTTP status code returned by the server.
        :param headers: A dictionary of headers returned by the server.
        :param url: URL of the retrieved resource represented by this Response.
        """
        self.fp = fp
        self.read = fp.read
        self.code = code
        self.headers = headers
        self.url = url

    def close(self):
        """Close the response body file object.
        """
        self.read = None
        self.fp = None

    def info(self):
        """Returns the response headers.
        """
        return self.headers

    def geturl(self):
        """Returns the URL for the resource returned in this response.
        """
        return self.url


class HttpErrorHandler(url_request.HTTPDefaultErrorHandler):
    """
    A custom HTTP error handler.

    Used to return Response objects instead of raising an HTTPError exception.
    """

    def http_error_default(self, req, fp, code, msg, headers):
        """Default HTTP error handler.

        :param req: The original Request object.
        :param fp: The response body file object.
        :param code: The HTTP status code returned by the server.
        :param msg: The HTTP status message returned by the server.
        :param headers: The response headers.
        :returns: A new Response object.
        """
        return Response(fp, code, headers, req.get_full_url())


class RemoteConnection(object):
    """A connection with the Remote WebDriver server.

    Communicates with the server using the WebDriver wire protocol:
    https://github.com/SeleniumHQ/selenium/wiki/JsonWireProtocol"""

    _default_timeout = settings.get('QT4I_XCTAGENT_CMD_TIMEOUT', 90)
    if _default_timeout == -1:
        _default_timeout = socket._GLOBAL_DEFAULT_TIMEOUT
    _timeout = _default_timeout
#     _timeout = socket._GLOBAL_DEFAULT_TIMEOUT
    

    @classmethod
    def get_timeout(cls):
        """
        
        :returns: Timeout value in seconds for all http requests made to the Remote Connection
        """
        return None if cls._timeout == socket._GLOBAL_DEFAULT_TIMEOUT else cls._timeout

    @classmethod
    def set_global_timeout(cls, timeout):
        """Override the default timeout
 
        :param timeout: timeout value for http requests in seconds
        :type timeout: int
        """
        cls._timeout = timeout
        
    @classmethod
    def reset_global_timeout(cls):
        """Reset the http request timeout to socket._GLOBAL_DEFAULT_TIMEOUT
        """
#         cls._timeout = socket._GLOBAL_DEFAULT_TIMEOUT
        cls._timeout = cls._default_timeout
    
    def set_timeout(self, timeout):
        self._timeout = timeout
        
    def reset_timeout(self):
        self._timeout = self._default_timeout

    def __init__(self, remote_server_addr, keep_alive=False, resolve_ip=True, logger_name='RemoteConnection'):
        # Attempt to resolve the hostname and get an IP address.
        self.logger = logging.get_logger(logger_name)
        self.keep_alive = keep_alive
        parsed_url = parse.urlparse(remote_server_addr)
        addr = ""
        if parsed_url.hostname and resolve_ip:
            try:
                netloc = socket.gethostbyname(parsed_url.hostname)
                addr = netloc
                if parsed_url.port:
                    netloc += ':%d' % parsed_url.port
                if parsed_url.username:
                    auth = parsed_url.username
                    if parsed_url.password:
                        auth += ':%s' % parsed_url.password
                    netloc = '%s@%s' % (auth, netloc)
                remote_server_addr = parse.urlunparse(
                    (parsed_url.scheme, netloc, parsed_url.path,
                     parsed_url.params, parsed_url.query, parsed_url.fragment))
            except socket.gaierror:
                self.logger.info('Could not get IP address for host: %s' % parsed_url.hostname)

        self._url = remote_server_addr
        if keep_alive:
            self._conn = httplib.HTTPConnection(
                str(addr), str(parsed_url.port), timeout=self._timeout)

        self._commands = {
            Command.HEALTH: ('GET', '/health'),
            Command.DEACTIVATE_APP: ('POST', '/session/$sessionId/wda/deactivateApp'),
            Command.STATUS: ('GET', '/status'),
            Command.SESSION: ('POST', '/session'),
            Command.GET_ALL_SESSIONS: ('GET', '/sessions'),
            Command.QUIT: ('DELETE', '/session/$sessionId'),
            Command.GET: ('POST', '/session/$sessionId/url'),
            Command.EXECUTE_SCRIPT: ('POST', '/session/$sessionId/execute'),
            Command.GET_CURRENT_URL: ('GET', '/session/$sessionId/url'),
            Command.GET_TITLE: ('GET', '/session/$sessionId/title'),
            Command.GET_PAGE_SOURCE: ('GET', '/session/$sessionId/source'),
            Command.SCREENSHOT: ('GET', '/screenshot'),
            Command.ELEMENT_SCREENSHOT: ('GET', '/session/$sessionId/screenshot/$id'),
            Command.FIND_ELEMENT: ('POST', '/session/$sessionId/element'),
            Command.FIND_ELEMENTS: ('POST', '/session/$sessionId/elements'),
            Command.GET_ACTIVE_ELEMENT:
                ('POST', '/session/$sessionId/element/active'),
            Command.FIND_CHILD_ELEMENT:
                ('POST', '/session/$sessionId/element/$id/element'),
            Command.FIND_CHILD_ELEMENTS:
                ('POST', '/session/$sessionId/element/$id/elements'),
            Command.CLICK_ELEMENT: ('POST', '/session/$sessionId/element/$id/click'),
            Command.CLEAR_ELEMENT: ('POST', '/session/$sessionId/element/$id/clear'),
            Command.SUBMIT_ELEMENT: ('POST', '/session/$sessionId/element/$id/submit'),
            Command.GET_ELEMENT_TREE: ('GET', '/session/$sessionId/tree'),
            Command.GET_ELEMENT_TEXT: ('GET', '/session/$sessionId/element/$id/text'),
            Command.SEND_KEYS_TO_ELEMENT:
                ('POST', '/session/$sessionId/element/$id/value'),
            Command.SEND_KEYS_TO_ACTIVE_ELEMENT:
                ('POST', '/session/$sessionId/keys'),
            Command.GET_ELEMENT_VALUE:
                ('GET', '/session/$sessionId/element/$id/value'),
            Command.GET_ELEMENT_TAG_NAME:
                ('GET', '/session/$sessionId/element/$id/name'),
            Command.IS_ELEMENT_SELECTED:
                ('GET', '/session/$sessionId/element/$id/selected'),
            Command.SET_ELEMENT_SELECTED:
                ('POST', '/session/$sessionId/element/$id/selected'),
            Command.IS_ELEMENT_ENABLED:
                ('GET', '/session/$sessionId/element/$id/enabled'),
            Command.IS_ELEMENT_DISPLAYED:
                ('GET', '/session/$sessionId/element/$id/displayed'),
            Command.GET_ELEMENT_LOCATION:
                ('GET', '/session/$sessionId/element/$id/location'),
            Command.GET_ELEMENT_LOCATION_ONCE_SCROLLED_INTO_VIEW:
                ('GET', '/session/$sessionId/element/$id/location_in_view'),
            Command.GET_ELEMENT_SIZE:
                ('GET', '/session/$sessionId/element/$id/size'),
            Command.GET_ELEMENT_RECT:
                ('GET', '/session/$sessionId/element/$id/rect'),
            Command.GET_ELEMENT_ATTRIBUTE:
                ('GET', '/session/$sessionId/element/$id/attribute/$name'),
            Command.ELEMENT_EQUALS:
                ('GET', '/session/$sessionId/element/$id/equals/$other'),
            Command.SWITCH_TO_FRAME: ('POST', '/session/$sessionId/frame'),
            Command.SWITCH_TO_PARENT_FRAME: ('POST', '/session/$sessionId/frame/parent'),
            Command.SWITCH_TO_WINDOW: ('POST', '/session/$sessionId/window'),
            Command.CLOSE: ('DELETE', '/session/$sessionId/window'),
            Command.IMPLICIT_WAIT:
                ('POST', '/session/$sessionId/timeouts/implicit_wait'),
            Command.EXECUTE_ASYNC_SCRIPT: ('POST', '/session/$sessionId/execute_async'),
            Command.SET_SCRIPT_TIMEOUT:
                ('POST', '/session/$sessionId/timeouts/async_script'),
            Command.SET_TIMEOUTS:
                ('POST', '/session/$sessionId/timeouts'),
            Command.DISMISS_ALERT:
                ('POST', '/session/$sessionId/dismiss_alert'),
            Command.ACCEPT_ALERT:
                ('POST', '/session/$sessionId/accept_alert'),
            Command.SET_ALERT_VALUE:
                ('POST', '/session/$sessionId/alert_text'),
            Command.GET_ALERT_TEXT:
                ('GET', '/session/$sessionId/alert_text'),
            Command.SET_ALERT_CREDENTIALS:
                ('POST', '/session/$sessionId/alert/credentials'),
            Command.CLICK:
                ('POST', '/session/$sessionId/click'),
            Command.DOUBLE_CLICK:
                ('POST', '/session/$sessionId/doubleclick'),
            Command.MOUSE_DOWN:
                ('POST', '/session/$sessionId/buttondown'),
            Command.MOUSE_UP:
                ('POST', '/session/$sessionId/buttonup'),
            Command.MOVE_TO:
                ('POST', '/session/$sessionId/moveto'),
            Command.GET_WINDOW_SIZE:
                ('GET', '/session/$sessionId/window/size'),
            Command.SET_WINDOW_SIZE:
                ('POST', '/session/$sessionId/window/$windowHandle/size'),
            Command.GET_WINDOW_POSITION:
                ('GET', '/session/$sessionId/window/$windowHandle/position'),
            Command.SET_WINDOW_POSITION:
                ('POST', '/session/$sessionId/window/$windowHandle/position'),
            Command.MAXIMIZE_WINDOW:
                ('POST', '/session/$sessionId/window/$windowHandle/maximize'),
            Command.SET_SCREEN_ORIENTATION:
                ('POST', '/session/$sessionId/orientation'),
            Command.GET_SCREEN_ORIENTATION:
                ('GET', '/session/$sessionId/orientation'),
            Command.SINGLE_TAP:
                ('POST', '/session/$sessionId/touch/click'),
            Command.TOUCH_DOWN:
                ('POST', '/session/$sessionId/touch/down'),
            Command.TOUCH_UP:
                ('POST', '/session/$sessionId/touch/up'),
            Command.TOUCH_MOVE:
                ('POST', '/session/$sessionId/touch/move'),
            Command.TOUCH_SCROLL:
                ('POST', '/session/$sessionId/touch/scroll'),
            Command.DOUBLE_TAP:
                ('POST', '/session/$sessionId/touch/doubleclick'),
            Command.LONG_PRESS:
                ('POST', '/session/$sessionId/touch/longclick'),
            Command.FLICK:
                ('POST', '/session/$sessionId/touch/flick'),
            Command.EXECUTE_SQL:
                ('POST', '/session/$sessionId/execute_sql'),
            Command.GET_LOCATION:
                ('GET', '/session/$sessionId/location'),
            Command.SET_LOCATION:
                ('POST', '/session/$sessionId/location'),
            Command.GET_APP_CACHE:
                ('GET', '/session/$sessionId/application_cache'),
            Command.GET_APP_CACHE_STATUS:
                ('GET', '/session/$sessionId/application_cache/status'),
            Command.CLEAR_APP_CACHE:
                ('DELETE', '/session/$sessionId/application_cache/clear'),
            Command.GET_NETWORK_CONNECTION:
                ('GET', '/session/$sessionId/network_connection'),
            Command.SET_NETWORK_CONNECTION:
                ('POST', '/session/$sessionId/network_connection'),
            Command.GET_LOCAL_STORAGE_ITEM:
                ('GET', '/session/$sessionId/local_storage/key/$key'),
            Command.REMOVE_LOCAL_STORAGE_ITEM:
                ('DELETE', '/session/$sessionId/local_storage/key/$key'),
            Command.GET_LOCAL_STORAGE_KEYS:
                ('GET', '/session/$sessionId/local_storage'),
            Command.SET_LOCAL_STORAGE_ITEM:
                ('POST', '/session/$sessionId/local_storage'),
            Command.CLEAR_LOCAL_STORAGE:
                ('DELETE', '/session/$sessionId/local_storage'),
            Command.GET_LOCAL_STORAGE_SIZE:
                ('GET', '/session/$sessionId/local_storage/size'),
            Command.GET_SESSION_STORAGE_ITEM:
                ('GET', '/session/$sessionId/session_storage/key/$key'),
            Command.REMOVE_SESSION_STORAGE_ITEM:
                ('DELETE', '/session/$sessionId/session_storage/key/$key'),
            Command.GET_SESSION_STORAGE_KEYS:
                ('GET', '/session/$sessionId/session_storage'),
            Command.SET_SESSION_STORAGE_ITEM:
                ('POST', '/session/$sessionId/session_storage'),
            Command.CLEAR_SESSION_STORAGE:
                ('DELETE', '/session/$sessionId/session_storage'),
            Command.GET_SESSION_STORAGE_SIZE:
                ('GET', '/session/$sessionId/session_storage/size'),
            Command.CURRENT_CONTEXT_HANDLE:
                ('GET', '/session/$sessionId/context'),
            Command.CONTEXT_HANDLES:
                ('GET', '/session/$sessionId/contexts'),
            Command.SWITCH_TO_CONTEXT:
                ('POST', '/session/$sessionId/context'),
            Command.QTA_FIND_ELEMENTS:
                ('POST', '/session/$sessionId/qta/element/$id/elements'),
            Command.QTA_DEVICE_CLICK:
                ('POST', '/session/$sessionId/qta/click'),
            Command.QTA_ELEMENT_CLICK:
                ('POST', '/session/$sessionId/qta/element/$id/click'),
            Command.QTA_DEVICE_DOUBLE_CLICK:
                ('POST', '/session/$sessionId/qta/doubleclick'),
            Command.QTA_ELEMENT_DOUBLE_CLICK:
                ('POST', '/session/$sessionId/wda/element/$id/doubleTap'),
            Command.QTA_DEVICE_LONG_CLICK:
                ('POST', '/session/$sessionId/qta/longclick'),
            Command.QTA_ELEMENT_LONG_CLICK:
                ('POST', '/session/$sessionId/wda/element/$id/touchAndHold'),
            Command.QTA_DEVICE_SENDKEYS:
                ('POST', '/session/$sessionId/qta/sendkeys'),
            Command.QTA_ELEMENT_SENDKEYS:
                ('POST', '/session/$sessionId/qta/element/$id/sendkeys'),
            Command.QTA_DEVICE_DRAG:
                ('POST', '/session/$sessionId/qta/drag'),
            Command.QTA_ELEMENT_DRAG:
                ('POST', '/session/$sessionId/qta/element/$id/drag'),
            Command.QTA_GET_PARENT_ELEMENT:
                ('POST', '/session/$sessionId/qta/element/$id/parent'),
            Command.QTA_GET_CHILDREN_ELEMENTS:
                ('POST', '/session/$sessionId/qta/element/$id/children'),
            Command.QTA_SCROLL_TO_VISIBLE:
                ('POST', '/session/$sessionId/qta/element/$id/scroll'),
            Command.QTA_DRAG_TO_VALUE:
                ('POST', '/session/$sessionId/qta/element/$id/slider'),
            Command.QTA_GET_ELEMENT_ATTRS:
                ('POST', '/session/$sessionId/qta/element/$id/attrs'),
            Command.QTA_ALERT_RULES_UPDATE:
                ('POST', '/qta/alertrules/update'),
            Command.QTA_ELEMENT_SETVALUE:
                ('POST', '/session/$sessionId/qta/element/$id/value'),
            Command.QTA_ELEMENT_TREE:
                ('POST', '/session/$sessionId/qta/element/$id/tree'),
            Command.QTA_STOP_AGENT:
                ('POST', '/qta/stop'),
            Command.QTA_GET_FOREGROUND_APP_NAME:
                ('GET', '/qta/appName'),
            Command.QTA_GET_FOREGROUND_APP_PID:
                ('GET', '/qta/appPid'),
            Command.QTA_GET_ELEMENT_WIN_INFO:
                ('GET', '/session/$sessionId/qta/element/$id/wininfo'),
            Command.QTA_ALERT_DISMISS:
                ('POST', '/qta/alert/dismiss'),
            Command.QTA_DEVICE_LOCK:
                ('POST', '/qta/device/lock'),    
            Command.QTA_DEVICE_UNLOCK:
                ('POST', '/qta/device/unlock'),  
            Command.QTA_SANDBOX_LIST:
                ('POST', '/qta/sandbox/list'),
            Command.QTA_SANDBOX_REMOVE:
                ('POST', '/qta/sandbox/remove'),
            Command.QTA_ALBUM_UPLOAD:
                ('POST', '/qta/album/upload'),
            Command.QTA_STUB_CALL:
                ('POST', '/qta/stub'),
            Command.QTA_DEVICE_VOLUME:
                ('POST', '/qta/device/volume'),     
            Command.QTA_DEVICE_SIRI:
                ('POST', '/qta/device/siri'),       
            Command.QTA_DEVICE_SCREEN_DIRECTION:
                ('POST', '/qta/device/screenDirection'), 
            Command.QTA_DEVICE_DETAIL_INFO:
                ('POST', '/qta/device/detailInfo'),
            Command.QTA_WHEEL_SELECT:
                ('POST', '/session/$sessionId/qta/element/$id/wheel/select'),
            Command.FORCE_TOUCH:
                ('POST', '/session/$sessionId/wda/element/forceTouch/$id'),
        }

    def execute(self, command, params):
        """Send a command to the remote server.
        Any path subtitutions required for the URL mapped to the command should be
        included in the command parameters.

        :param command: A string specifying the command to execute.
        :type command: str
        :param params: A dictionary of named parameters to send with the command as its JSON payload.
        :type params: dict
        :returns: dict 
        """
        command_info = self._commands[command]
        assert command_info is not None, 'Unrecognised command %s' % command
        data = json.dumps(params)
        path = string.Template(command_info[1]).substitute(params)
        url = '%s%s' % (self._url, path)
        return self._request(command_info[0], url, body=data)

    def _request(self, method, url, body=None):
        """Send an HTTP request to the remote server.

        :param method: A string for the HTTP method to send the request with.
        :type method: str
        :param url: A string for the URL to send the request to.
        :type url: str
        :param body: A string for request body. Ignored unless method is POST or PUT.
        :type body: str
        :returns: A dictionary with the server's parsed JSON response.
        """
        self.logger.debug('%s %s %s' % (method, url, body))
        
        for _ in range(3):
            try:
                parsed_url = parse.urlparse(url)
        
                if self.keep_alive:
                    headers = {"Connection": 'keep-alive', method: parsed_url.path,
                               "User-Agent": "Python http auth",
                               "Content-type": "application/json;charset=\"UTF-8\"",
                               "Accept": "application/json"}
                    if parsed_url.username:
                        auth = base64.standard_b64encode(('%s:%s' %
                               (parsed_url.username, parsed_url.password)).encode('ascii')).decode('ascii').replace('\n', '')
                        headers["Authorization"] = "Basic %s" % auth
                    if body and method != 'POST' and method != 'PUT':
                        body = None
                    try:
                        self._conn.request(method, parsed_url.path, body, headers)
                        resp = self._conn.getresponse()
                    except (httplib.HTTPException, socket.error):
                        self._conn.close()
                        raise
        
                    statuscode = resp.status
                else:
                    password_manager = None
                    if parsed_url.username:
                        netloc = parsed_url.hostname
                        if parsed_url.port:
                            netloc += ":%s" % parsed_url.port
                        cleaned_url = parse.urlunparse((parsed_url.scheme,
                                                           netloc,
                                                           parsed_url.path,
                                                           parsed_url.params,
                                                           parsed_url.query,
                                                           parsed_url.fragment))
                        password_manager = url_request.HTTPPasswordMgrWithDefaultRealm()
                        password_manager.add_password(None,
                                                      "%s://%s" % (parsed_url.scheme, netloc),
                                                      parsed_url.username,
                                                      parsed_url.password)
                        request = Request(cleaned_url, data=body.encode('utf-8'), method=method)
                    else:
                        request = Request(url, data=body.encode('utf-8'), method=method)
        
                    request.add_header('Accept', 'application/json')
                    request.add_header('Content-Type', 'application/json;charset=UTF-8')
        
                    if password_manager:
                        opener = url_request.build_opener(url_request.HTTPRedirectHandler(),
                                                          HttpErrorHandler(),
                                                          url_request.HTTPBasicAuthHandler(password_manager),
                                                          url_request.ProxyHandler({}))
                    else:
                        opener = url_request.build_opener(url_request.HTTPRedirectHandler(),
                                                          HttpErrorHandler(),
                                                          url_request.ProxyHandler({}))
                    resp = opener.open(request, timeout=self._timeout)
                    statuscode = resp.code
                    if not hasattr(resp, 'getheader'):
                        if hasattr(resp.headers, 'getheader'):
                            resp.getheader = lambda x: resp.headers.getheader(x)
                        elif hasattr(resp.headers, 'get'):
                            resp.getheader = lambda x: resp.headers.get(x)
        
                data = resp.read()
                try:
                    if 300 <= statuscode < 304:
                        return self._request('GET', resp.getheader('location'))
                    body = data.decode('utf-8').replace('\x00', '').strip()
                    if 399 < statuscode < 500:
                        return {'status': statuscode, 'value': body}
                    content_type = []
                    if resp.getheader('Content-Type') is not None:
                        content_type = resp.getheader('Content-Type').split(';')
                    if not any([x.startswith('image/png') for x in content_type]):
                        try:
                            data = json.loads(body.strip())
                        except ValueError:
                            if 199 < statuscode < 300:
                                status = ErrorCode.SUCCESS
                            else:
                                status = ErrorCode.UNKNOWN_ERROR
                            return {'status': status, 'value': body.strip()}
        
                        assert type(data) is dict, (
                            'Invalid server response body: %s' % body)
                        # Some of the drivers incorrectly return a response
                        # with no 'value' field when they should return null.
                        if 'value' not in data:
                            data['value'] = None
                        return data
                    else:
                        data = {'status': 0, 'value': body.strip()}
                        return data
                finally:
                    self.logger.debug("Finished Request")
                    resp.close()
                
            except socket.timeout:
                self.logger.error('Remote Connection timeout')
                raise XCTestAgentTimeoutException('XCTestAgent response is timed out')
            except Exception as e:
                self.logger.error('Remote Connection:%s' % str(e))



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''设备型号与设备信息映射
'''

class DeviceProperty(object):
    
    mapping_simulator_table = {   
        #iPhone SE
        'iPhone SE,N/A,N/A':'A1662, A1723, A1724',
        
        #iPhone 6
        'iPhone 6,N/A,N/A':'A1549, A1586, A1589',
        
        #iPhone 6 Plus
        'iPhone 6 Plus,N/A,N/A':'A1522, A1524, A1593',
        
        #iPhone 6s
        'iPhone 6s,N/A,N/A':'A1633, A1688, A1691, A1700',
        
        #iPhone 6s Plus
        'iPhone 6s Plus,N/A,N/A':'A1634, A1687, A1690, A1699',
        
        #iPhone 7
        'iPhone 7,N/A,N/A':'A1660, A1779, A1780, A1778',
        
        #iPhone 7 Plus
        'iPhone 7 Plus,N/A,N/A':'A1661, A1785, A1786, A1784',
        
        #iPhone 8
        'iPhone 8,N/A,N/A':'A1863, A1906, A1907, A1905',
        
         #iPhone 8 Plus
        'iPhone 8 Plus,N/A,N/A':'A1864, A1898, A1899, A1897',
        
        #iPhone X
        'iPhone X,N/A,N/A':'A1865, A1902, A1901',
        }
    
    mapping_real_table = {   
        #iPhone SE
        'iPhone SE,Gold,16GB':'MLXH2, MLXW2, MLY12, MLY52, MLY92, MLXM2',
        'iPhone SE,Gold,32GB' : 'MP8D2, MP8R2, MP8H2, MP8M2, MP7V2, MP842',
        'iPhone SE,Gold,64GB' : 'MLXK2, MLXY2, MLY32, MLY72, MLYC2, MLXP2',
        'iPhone SE,Gold,128GB': 'MP952, MP9J2, MP992, MP9E2, MP802, MP882',
        
        'iPhone SE,Rose Gold,16GB' : 'MLXJ2, MLXX2, MLY22, MLY62, MLYA2, MLXN2',
        'iPhone SE,Rose Gold,32GB' : 'MP8E2, MP8T2, MP8J2, MP8N2, MP7W2, MP852',
        'iPhone SE,Rose Gold,64GB' : 'MLXL2, MLY02, MLY42, MLY82, MLYD2, MLXQ2',
        'iPhone SE,Rose Gold,128GB': 'MP962, MP9K2, MP9A2, MP9F2, MP812, MP892',
        
        'iPhone SE,Silver,16GB' : 'MLLM2, MLLV2, MLLX2, MLM02, MLM32, MLLP2',
        'iPhone SE,Silver,32GB' : 'MP8C2, MP8Q2, MP8G2, MP8L2, MP7U2, MP832',
        'iPhone SE,Silver,64GB' : 'MLM52, MLMC2, MLME2, MLMG2, MLMJ2, MLM72',
        'iPhone SE,Silver,128GB': 'MP942, MP9H2, MP982, MP9D2, MP7Y2, MP872',
        
        'iPhone SE,Space Gray,16GB' : 'MLLL2, MLLU2, MLLW2, MLLY2, MLM22, MLLN2',
        'iPhone SE,Space Gray,32GB' : 'MP8A2, MP8P2, MP8F2, MP8K2, MP7T2, MP822',
        'iPhone SE,Space Gray,64GB' : 'MLM42, MLM62, MLMA2, MLMD2, MLMF2, MLMH2',
        'iPhone SE,Space Gray,128GB': 'MP932, MP9G2, MP972, MP9C2, MP7X2, MP862',
        
        #iPhone 6
        'iPhone 6,Gold,16GB' : 'MG3D2, MG492, MG4Q2, MG562, MG5Y2, MG6C2',
        'iPhone 6,Gold,64GB' : 'MG3L2, MG4J2, MG502, MG652, MG6J2, MG842',
        'iPhone 6,Gold,128GB': 'MG3G2, MG4E2, MG4V2, MG622',
        
        'iPhone 6,Silver,16GB' : 'MG3C2, MG4P2, MG482, MG5X2, MG552, MG6A2, MG7W2',
        'iPhone 6,Silver,64GB' : 'MG3K2, MG4H2, MG4X2, MG5C2, MG6H2, MG642, MG822',
        'iPhone 6,Silver,128GB': 'MG3F2, MG4C2',
        
        'iPhone 6,Space Gray,16GB': 'MG3A2, MG4N2, MG472, MG5W2, MG542, MG692',
        'iPhone 6,Space Gray,64GB': 'MG3H2, MG4F2, MG4W2, MG5A2, MG6G2, MG632',
        'iPhone 6,Space Gray,128GB': 'MG3E2, MG4A2, MG4R2, MG602',
        
        #iPhone 6 Plus
        'iPhone 6 Plus,Gold,16GB' : 'MGAA2, MGAN2, MGC12, MGCM2, MGCX2',
        'iPhone 6 Plus,Gold,64GB' : 'MGAK2, MGAW2, MGC72',
        'iPhone 6 Plus,Gold,128GB': 'MGAF2, MGAR2, MGC42, MGCQ2',
        
        'iPhone 6 Plus,Silver,16GB' : 'MGA92, MGC92, MGAM2, MGC02, MGCL2',
        'iPhone 6 Plus,Silver,64GB' : 'MGAJ2, MGAV2, MGC62',
        'iPhone 6 Plus,Silver,128GB': 'MGAE2, MGAQ2, MGC32',
        
        'iPhone 6 Plus,Space Gray,16GB': 'MGA82, MGAL2, MGAX2, MGCK2',
        'iPhone 6 Plus,Space Gray,64GB': 'MGAH2, MGAU2, MGC52, MGCR2',
        'iPhone 6 Plus,Space Gray,128GB': 'MGAC2, MGAP2, MGC22, MGCN2',
        
        #iPhone 6s
        'iPhone 6s,Gold,16GB' : 'MKQL2, MKQ72, MKR12, MKRE2, MKRW2, MKT92, ML7E2',
        'iPhone 6s,Gold,32GB' : 'MN0P2, MN172, MN1K2, MN1U2, MN1Y2',
        'iPhone 6s,Gold,64GB' : 'MKQC2, MKQQ2, MKR52, MKRJ2, MKT12, MKTE2, ML7J2',
        'iPhone 6s,Gold,128GB': 'MKQG2, MKQV2, MKR92, MKRP2, MKT52, MKTJ2, ML7N2',
        
        'iPhone 6s,Rose Gold,16GB' : 'MKQM2, MKQ82, MKRF2, MKRX2, MKR22, MKTA2, ML7F2',
        'iPhone 6s,Rose Gold,32GB' : 'MN0V2, MN192, MN1L2, MN1V2, MN202',
        'iPhone 6s,Rose Gold,64GB' : 'MKQD2, MKQR2, MKR62, MKRK2, MKT22, MKTF2, ML7K2',
        'iPhone 6s,Rose Gold,128GB': 'MKQH2, MKQW2, MKRA2, MKRQ2, MKT62, MKTK2, ML7P2',
        
        'iPhone 6s,Silver,16GB' : 'MKQ62, MKQK2, MKQY2, MKRD2, MKRT2, MKT82, ML7D2, NKQJ2',
        'iPhone 6s,Silver,32GB' : 'MN0N2, MN162, MN1G2, MN1Q2, MN1X2',
        'iPhone 6s,Silver,64GB' : 'MKQA2, MKQP2, MKR42, MKRH2, MKT02, MKTD2, ML7H2',
        'iPhone 6s,Silver,128GB': 'MKQF2, MKQU2, MKR82, MKRM2, MKT42, MKTH2, ML7M2',
        
        'iPhone 6s,Space Gray,16GB': 'MKQ52, MKQJ2, MKQX2, MKRC2, MKRR2, MKT72, ML7C2',
        'iPhone 6s,Space Gray,32GB': 'MN0M2, MN132, MN1E2, MN1M2, MN1W2',
        'iPhone 6s,Space Gray,64GB': 'MKQN2, MKQ92, MKR32, MKRG2, MKRY2, MKTC2, ML7G2',
        'iPhone 6s,Space Gray,128GB': 'MKQE2, MKQT2, MKR72, MKRL2, MKT32, MKTG2, ML7L2',
        
        #iPhone 6s Plus
        'iPhone 6s Plus,Gold,16GB' : 'MKTN2, MKU32, MKUN2, MKV62, MKVQ2, MKW72, ML6D2',
        'iPhone 6s Plus,Gold,64GB' : 'MKTT2, MKU82, MKUV2, MKVD2, MKVX2, MKWD2, ML6H2',
        'iPhone 6s Plus,Gold,128GB': 'MKTX2, MKUF2, MKV12, MKVH2, MKW22, MKWH2, ML6M2',
        
        'iPhone 6s Plus,Rose Gold,16GB' : 'MKU52, ML6E2, MKTP2, MKUP2, MKV72, MKVU2, MKW82',
        'iPhone 6s Plus,Rose Gold,64GB' : 'MKU92, ML6J2, MKTU2, MKUW2, MKVE2, MKVY2, MKWE2',
        'iPhone 6s Plus,Rose Gold,128GB': 'MKUG2, MKV22, ML6N2, MKVJ2, MKTY2',
        
        'iPhone 6s Plus,Silver,16GB' : 'MKTM2, MKU22, MKUJ2, MKV52, MKVP2, MKW62, ML6C2',
        'iPhone 6s Plus,Silver,64GB' : 'MKTR2, MKU72, MKUU2, MKV92, MKVW2, MKWC2, ML6G2',
        'iPhone 6s Plus,Silver,128GB': 'MKTW2, MKUE2, MKUY2, MKVG2, MKW12, MKWG2, ML6L2',
        
        'iPhone 6s Plus,Space Gray,16GB': 'MKU12, ML6A2, MKV32, MKTL2, MKUH2, MKVN2, MKW52',
        'iPhone 6s Plus,Space Gray,64GB': 'MKU62, ML6F2, MKTQ2, MKUQ2, MKV82, MKVV2, MKW92, ML642',
        'iPhone 6s Plus,Space Gray,128GB': 'MKUD2, ML6K2, MKTV2, MKUX2, MKVF2, MKW02, MKWF2',
        
        #iPhone 7
        'iPhone 7,black,32GB' : 'MN8G2, MNAC2, MNAY2, MNCE2, MNGQ2, MN9D2, MN9U2, MN8X2',
        'iPhone 7,black,128GB' : 'MN8L2, MNAJ2, MNC32, MNCK2, MNGX2, MN9H2, MN9Y2, MN922',
        'iPhone 7,black,256GB': 'MN8R2, MNAQ2, MNC82, MNCQ2, MN9N2, MNA62, MN972',
        
        'iPhone 7,Gold,32GB' : 'MN8J2, MNAE2, MNC12, MNCG2, MN9F2, MN9W2, MN902',
        'iPhone 7,Gold,128GB' : 'MN8N2, MNAL2, MNC52, MNCM2, MNH02, MN9K2, MNA32, MN942',
        'iPhone 7,Gold,256GB': 'MN8U2, MNAV2, MNCA2, MNCT2, MN9Q2, MNA82',
        
        'iPhone 7,Jet Black,128GB': 'MN8Q2, MNAP2, MNC72, MNCP2, MN9M2, MNA52, MN962',
        'iPhone 7,Jet Black,256GB': 'MN8W2, MNAX2, MNCD2, MNCV2, MN9T2, MNAA2, MN9C2',
        
        'iPhone 7,Red,128GB' : 'MPRV2, MPRT2, MPRH2, MPRL2, MPRN2, MPRQ2',
        'iPhone 7,Red,256GB' : 'MPRW2, MPRU2, MPRJ2, MPRM2, MPRP2, MPRR2',
        
        'iPhone 7,Rose Gold,32GB' : 'MN8K2, MNAF2, MNC22, MNCJ2, MN9G2, MN9X2, MN912',
        'iPhone 7,Rose Gold,128GB' : 'MN8P2, MNAM2, MNC62, MNH12, MNCN2, MN9L2, MNA42, MN952',
        'iPhone 7,Rose Gold,256GB': 'MN8V2, MNAW2, MNCC2, MNCU2, MN9R2, MNA92',
        
        'iPhone 7,Silver,32GB' : 'MN8H2, MNAD2, MNC02, MNCF2, MN9E2, MN9V2',
        'iPhone 7,Silver,128GB' : 'MN8M2, MNAK2, MNC42, MNCL2, MN9J2, MNA02',
        'iPhone 7,Silver,256GB': 'MN8T2, MNAU2, MNC92, MNCR2, MN9P2, MNA72',
        
        #iPhone 7 Plus
        'iPhone 7 Plus,black,32GB' : 'MNQH2, MNR12, MNR52, MNR92, MNRJ2, MNQR2, MNQW2, MNQM2',
        'iPhone 7 Plus,black,128GB' : 'MN482, MN5T2, MN642, MN6F2, MNFP2, MN522, MN5G2, MN4M2',
        'iPhone 7 Plus,black,256GB': 'MN4E2, MN5Y2, MN692, MN6L2, MNFV2, MN592, MN5M2, MN4W2',
        
        'iPhone 7 Plus,Gold,32GB' : 'MNQK2, MNR32, MNR72, MNRC2, MNRL2, MNQU2, MNQY2, MNQP2',
        'iPhone 7 Plus,Gold,128GB' : 'MN4A2, MN5V2, MN662, MN6H2, MNFR2, MN552, MN5J2, MN4Q2',
        'iPhone 7 Plus,Gold,256GB': 'MN4J2, MN612, MN6C2, MN6N2, MN5D2, MN5P2, MN4Y2',
        
        'iPhone 7 Plus,Jet Black,128GB': 'MN4D2, MN5X2, MN682, MN6K2, MN572, MN5L2, MN4V2',
        'iPhone 7 Plus,Jet Black,256GB': 'MN4L2, MN632, MN6E2, MN6Q2, MN5F2, MN5R2',
        
        'iPhone 7 Plus,Red,128GB' : 'MPR12, MPR02, MPQV2, MPQW2, MPQW2, MPQY2',
        'iPhone 7 Plus,Red,256GB' : 'MPRA2, MPR92, MPR52, MPR62, MPR72, MPR82',
        
        'iPhone 7 Plus,Rose Gold,32GB' : 'MNQL2, MNR42, MNR82, MNRD2, MNRM2, MNQV2, MNR02, MNQQ2',
        'iPhone 7 Plus,Rose Gold,128GB' : 'MN4C2, MN5W2, MN672, MN6J2, MNFT2, MN562, MN5K2, MN4U2',
        'iPhone 7 Plus,Rose Gold,256GB': 'MN4K2, MN622, MN6D2, MN6P2, MNFY2, MN5E2, MN5Q2',
        
        'iPhone 7 Plus,Silver,32GB' : 'MNQJ2, MNR22, MNR62, MNRA2, MNRK2, MNQT2, MNQX2, MNQN2',
        'iPhone 7 Plus,Silver,128GB' : 'MN492, MN5U2, MN652, MN6G2, MNFQ2, MN532, MN5H2, MN4P2',
        'iPhone 7 Plus,Silver,256GB': 'MN4F2, MN602, MN6A2, MN6M2, MN5C2, MN5N2, MN4X2, MN502',
        
        #iPhone8p
        'iPhone 8,Gold,64GB' : 'MQ6M2, MQ742, MQ772, MQ6J2, MQ6X2, MQ712',
        'iPhone 8,Gold,246GB': 'MQ7H2, MQ802, MQ832, MQ7E2, MQ7T2, MQ7W2',
        
        'iPhone 8,Silver,64GB' : 'MQ6L2, MQ732, MQ762, MQ6H2, MQ6W2, MQ702',
        'iPhone 8,Silver,256GB': 'MQ7G2, MQ7Y2, MQ822, MQ7D2, MQ7R2, MQ7V2',
        
        'iPhone 8,Space Gray,64GB': 'MQ6K2, MQ722, MQ752, MQ6G2, MQ6V2, MQ6Y2',
        'iPhone 8,Space Gray,256GB': 'MQ7F2, MQ7X2, MQ812, MQ7C2, MQ7Q2, MQ7U2',
        
        #iPhone8p
        'iPhone 8 Plus,Gold,64GB' : 'MQ8F2, MQ9F2, MQ982, MQ8N2, MQ8V2, MQ922',
        'iPhone 8 Plus,Gold,246GB': 'MQ8J2, MQ9C2, MQ9J2, MQ8R2, MQ8Y2, MQ952',
        
        'iPhone 8 Plus,Silver,64GB' : 'MQ8E2, MQ9E2, MQ972, MQ8M2, MQ8U2, MQ912',
        'iPhone 8 Plus,Silver,256GB': 'MQ8H2, MQ9A2, MQ9H2, MQ8Q2, MQ8X2, MQ942',
        
        'iPhone 8 Plus,Space Gray,64GB': 'MQ8D2, MQ9D2, MQ962, MQ8L2, MQ8T2, MQ902',
        'iPhone 8 Plus,Space Gray,256GB': 'MQ8G2, MQ9G2, MQ992, MQ8P2, MQ8W2, MQ932',
        
        #iPhoneX     
        'iPhone X,Silver,64GB' : 'MQCT2, MQCL2, MQAY2, MQA62, MQAK2, MQAR2, MQAD2',
        'iPhone X,Silver,256GB': 'MQCW2, MQCP2, MQC22, MQA92, MQAN2, MQAV2, MQAG2',
        
        'iPhone X,Space Gray,64GB': 'MQCR2, MQCK2, MQAX2, MQA52, MQAJ2, MQAQ2, MQAC2',
        'iPhone X,Space Gray,256GB': 'MQCV2, MQCN2, MQC12, MQA82, MQAM2, MQAU2, MQAF2',
        }

    @classmethod
    def get_property(cls, device_type, is_simulator):
        '''根据device_type,获取设备详细信息
        
        :param device_type : 设备型号
        :type device_type : str
        :param is_simulator : 模拟器
        :type is_simulator : bool
        
        :returns: attr - 设备属性字典
        '''
        attr = ''
        if is_simulator :
            dic = cls.mapping_simulator_table
        else :
            dic = cls.mapping_real_table
        for i in range(len(dic)):
            if attr != '' :
                break
            key = dic.keys()[i]
            value = dic[key]
            l = value.split(',')
            for i in l:
                if i.strip() == device_type:
                    attr = key
                    break
        if attr != '' :
            l = attr.split(',')
            attr = {
                    'model' : l[0],
                    'color' : l[1],
                    'storage' : l[2], 
                    }
        else :
            attr = {
                    'model' : 'Unknown',
                    'color' : 'N/A',
                    'storage' : 'N/A', 
                    }
        return attr


# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''XCUITest API driver
'''

from __future__ import absolute_import, print_function

import os
import time
import select
import threading
import pkg_resources
import traceback
import shutil
import subprocess
from six import with_metaclass

from qt4i.driver.tools.sched import PortManager
from qt4i.driver.tools import logger, dt
from qt4i.driver.util import Task
from qt4i.driver.util import ThreadTask
from qt4i.driver.util import Process
from qt4i.driver.xctest.webdriverclient.command import Command
from qt4i.driver.xctest.webdriverclient.errorhandler import ErrorHandler
from qt4i.driver.xctest.webdriverclient.exceptions import WebDriverException
from qt4i.driver.xctest.webdriverclient.exceptions import XCTestAgentDeadException
from qt4i.driver.xctest.webdriverclient.exceptions import XCTestAgentTimeoutException
from qt4i.driver.xctest.webdriverclient.exceptions import ApplicationCrashedException
from qt4i.driver.xctest.webdriverclient.remote_connection import RemoteConnection
from pymobiledevice.usbmux.tcprelay import ThreadedTCPServer, TCPRelay
from testbase.conf import settings
from testbase.util import Singleton

CurrentDirectoryPath = os.path.abspath(os.path.dirname(__file__))

MAX_PORT_NUM = 50
DEFAULT_IP = '127.0.0.1'
DEFAULT_PORT = 8100
AGENT_STORE_PATH = '/cores/xctestagent'


class EnumDevice():
    '''设备类型
    '''
    Device = "iOS"
    Simulator = "iOS Simulator"
    
class AgentStartError(Exception):
    '''Agent Start Error
    '''
    pass


class XCUITestAgentManager(with_metaclass(Singleton, object)):
    '''XCUITest Agent Manager 单例类
    '''
    
    _agents = {}  # 维护Agents的实例，key为设备的udid，value为Agent实例
    _lock = threading.Lock()
    
    @classmethod
    def stop_all_agents(cls):
        '''关闭全部Agent
        '''
        for device_id in cls._agents:
            cls._agents[device_id].stop()
            PortManager.del_port('agent', device_id)
            logger.get_logger("xctest_%s" % device_id).info('stop_agent')
        cls._agents = {}

    def start_agent(self, device_id, server_ip=DEFAULT_IP, server_port=DEFAULT_PORT, keep_alive=False, retry=3, timeout=60):
        '''启动Agent并返回
        
        :param device_id: 设备ID
        :type device_id: str
        :param server_ip: 设备IP地址
        :type server_ip: str
        :param server_port: 设备端口号
        :type server_port: int
        :param keep_alive: HTTP远程连接是否使用keep-alive，默认为False
        :type keep_alive: boolean
        :param retry: 启动尝试次数，默认3次
        :type retry: int
        :param timeout: 单次启动超时 (秒)
        :type timeout: int
        :returns: XCUITestAgent
        '''
        self.log = logger.get_logger("xctest_%s" % device_id)
        if device_id not in self._agents:
            try:
                PortManager.set_port('agent', device_id, server_port)
                server_port = PortManager.get_port('agent', device_id)
                self.log.info('start_agent, port: %d' %server_port)
                with self._lock:
                    self._agents[device_id] = XCUITestAgent(device_id, server_ip, server_port, keep_alive, retry, timeout)
            except:
                self.log.exception('start_agent')
                PortManager.del_port('agent', device_id)
                raise
        return self._agents[device_id]
        
    def restart_agent(self, device_id):
        '''重启Agent
        '''
        if device_id in self._agents:
            self._agents[device_id].stop()
            self._agents[device_id].start()
        else:
            self.start_agent(device_id)
        
    def stop_agent(self, device_id):
        '''关闭Agent
        '''
        if device_id in self._agents:
            self._agents[device_id].stop()
            PortManager.del_port('agent', device_id)
            self._agents.pop(device_id)
    
    def get_agent(self, device_id):
        '''根据udid获取Agent对象
        
        :param device_id: 设备udid
        :type device_id: str
        :returns: XCUITestAgent
        '''
        if device_id in self._agents:
            return self._agents[device_id]
        else:
            return self.start_agent(device_id)
    
    @property
    def agents(self):
        '''获取Agent列表
        
        :returns: list
        '''
        return self._agents.values()
    
    def has_agent(self, device_id):
        '''查询设备是否存在对应的agent
        
        :returns: boolean
        '''
        return device_id in self._agents


class XCUITestAgent(object):
    '''XCUITest Agent
    '''
    _process = None # Agent进程
    _relay_thread = None # 端口映射线程
    _agent_cmd = ''
    CommandTimeout = settings.get('QT4I_START_APP_TIMEOUT', 60)
    PythonPath = Task("which python").execute()
    
    LOG_FILTERED_METHODS = [
                            Command.SCREENSHOT, 
                            Command.QTA_STOP_AGENT,
                            Command.GET_PAGE_SOURCE,
                            Command.GET_ELEMENT_TREE,
                            Command.QTA_ELEMENT_TREE,
                            ]

    def __init__(self, device_id, server_ip=DEFAULT_IP, server_port=DEFAULT_PORT, keep_alive=False, retry=3, timeout=60):
        self.log_name = "xctest_%s" % device_id
        self.log = logger.get_logger(self.log_name)
        
        self.version = '0.0.0'
        version_file = os.path.join(os.path.expanduser('~'), 'XCTestAgent', 'version.txt')
        if os.path.exists(version_file):
            with open(version_file, 'r') as f:
                self.version = f.read()
        
        # XCTestAgent工程默认路径：~/XCTestAent/XCTestAgent.xcodeproj
        self.XCTestAgentPath = os.path.join(os.path.expanduser('~'), "XCTestAgent", "XCTestAgent.xcodeproj")
        if not os.path.exists(self.XCTestAgentPath):
            msg = 'XCTestAgent does not exist. Please use command line: "python manage.py qt4i.setup" to setup XCTestAgent.xcodeproj'
            self.log.error(msg)
            raise Exception(msg)
            
        self.udid = device_id
        self._server_ip = server_ip
        self._server_port = server_port
        self._server_url = 'http://%s:%d' % (server_ip, server_port)
        self.type = EnumDevice.Simulator if dt.DT().is_simulator(device_id) else EnumDevice.Device
        self._is_remote = True
        self.session_id = None
        self.crash_flag = False
        self.capabilities = {}
        self.mjpeg_port = 9100 + self._server_port - 8100
        self.stub_remote_port = 18123
        self.stub_port = 18123 + self._server_port - 8100
        self.stub_server_url = 'http://%s:%d' % (DEFAULT_IP, self.stub_port)
        self.stub_client = RemoteConnection(self.stub_server_url, keep_alive=False, logger_name=self.log_name)
        self.stub_client.set_timeout(0.01)
        self.error_handler = ErrorHandler()
        self._command_executor = RemoteConnection(self._server_url, keep_alive=keep_alive, logger_name=self.log_name)
        self._is_relay_quit = threading.Event()
        self.start(retry, timeout)
    
    def _tcp_relay(self):
        '''将设备8100端口映射到本地8100端口，虚拟机不需要映射
        '''
        if self.type == EnumDevice.Simulator:
            return
        if not self._relay_thread and self._server_ip == DEFAULT_IP:
            self._relay_error = None
            self._relay_thread = threading.Thread(target=self._forward_ports)
            self._relay_thread.daemon = True
            self.log.info("Start TCPRelay Thread")
            self._relay_thread.start()
            time.sleep(1)
            if self._relay_error:
                raise Exception(self._relay_error)
            if not self._relay_thread.is_alive():
                raise RuntimeError('_relay_thread failed to start')
            
    def _forward_ports(self):
        '''iOS真机设备的端口转发
        '''
        try:
            agent_port_pair = '%d:%d' % (8100, self._server_port)
            stub_port_pair = '%d:%d' %(self.stub_remote_port, self.stub_port)
            mjpeg_port_pair = '%d:%d' %(9100, self.mjpeg_port)
            pair_ports = [agent_port_pair, stub_port_pair, mjpeg_port_pair] # 端口对的数组，每对端口中前一个代表远程端口，后一个代表本地端口，例如：["8100:8100", "8200:8200"]
            self._tcp_servers=[]
            serverclass = ThreadedTCPServer
            for pair_port in pair_ports:
                rport, lport = pair_port.split(":")
                rport = int(rport)
                lport = int(lport)
                self.log.info("Forwarding local port %d to remote port %d" % (lport, rport))
                server = serverclass(("localhost", lport), TCPRelay)
                server.rport = rport
                server.bufsize = 128
                server.udid = self.udid
                self._tcp_servers.append(server)
                
            self._quit_relay_thread = False # 退出端口转发线程的标识位
            self._is_relay_quit.clear()
            while not self._quit_relay_thread:
                try:
                    rl, wl, xl = select.select(self._tcp_servers, [], []) #@UnusedVariable
                    for server in rl:
                        server.handle_request()
                except:
                    self.log.exception('Forward ports')
                    self._quit_relay_thread = True
            self.log.info("Forward ports: successfully quit select")
            for server in self._tcp_servers:
                server.server_close()
            self._is_relay_quit.set()
            self.log.info("Forward ports: successfully close socket")
            
        except Exception:
            self._relay_error = traceback.format_exc()
            self.log.exception('forward ports')

    def _build_agent_for_simulator(self, xcode_version):
        agent_dir = os.path.join(AGENT_STORE_PATH, xcode_version)
        agent_file = os.path.join(agent_dir, 'XCTestAgent-Runner.app')
        if not os.path.exists(agent_file):
            if not os.path.exists(agent_dir):
                os.makedirs(agent_dir)
            xctest_root = pkg_resources.resource_filename("qt4i", "driver/xctest/xctestproj") #@UndefinedVariable
            xctestproj = os.path.join(xctest_root, 'XCTestAgent','XCTestAgent.xcodeproj')
            build_for_test_cmd = ' '.join(['xcodebuild',
                'build-for-testing',
                '-project',
                '"%s"' % xctestproj,
                '-scheme',
                'XCTestAgent',
                '-destination',
                '"platform=%s,id=%s"' % (self.type, self.udid),
                'CONFIGURATION_BUILD_DIR=%s' % agent_dir])
            subprocess.call(build_for_test_cmd, shell=True)
            self.log.debug("build XCTestAgent:%s" % build_for_test_cmd)
        return agent_dir

    def _get_xcodebuild_agent_cmd_for_simulator(self, xcode_version):
        template_root = pkg_resources.resource_filename("qt4i", "driver/xctest/bin") #@UndefinedVariable
        xctestfile_template = os.path.join(template_root, "iphonesimulator-%s.xctestrun" % xcode_version)
        agent_dir = self._build_agent_for_simulator(xcode_version)
        xctestfile = os.path.join(agent_dir, "%s.xctestrun" % self.udid)
        if not os.path.isfile(xctestfile):
            shutil.copyfile(xctestfile_template, xctestfile)
        cmd = "/usr/libexec/PlistBuddy -c 'Set :XCTestAgent:EnvironmentVariables:USE_PORT %s' %s" % (self._server_port, xctestfile)
        subprocess.call(cmd, shell=True)
        derived_data_path = '/tmp/xctlog/%s' % self.udid
        if not os.path.exists(derived_data_path):
            os.makedirs(derived_data_path)
        return ' '.join(['xcodebuild',
                'test-without-building',
                '-xctestrun',
                xctestfile,
                '-destination',
                '"platform=%s,id=%s"' % (self.type, self.udid),
                '-derivedDataPath',
                derived_data_path])

    def _get_fbsimctl_agent_cmd_for_simulator(self, xcode_version):
        agent_dir = self._build_agent_for_simulator(xcode_version)
        xctest_path = os.path.join(agent_dir, 'XCTestAgent-Runner.app/PlugIns/XCTestAgent.xctest') 
        attached_app = settings.get('QT4I_SIM_ATTACHED_APP', 'com.apple.reminders')
        return ' '.join([dt.DT().fbsimctl,
                self.udid,
                'launch_xctest',
                xctest_path,
                attached_app,
                '--port %s -- listen' % self._server_port])
        
    def _stdout_line_callback(self, line):
        '''标准输出回调函数
        '''
        self.log.debug(line.strip(' '))
    
    def _stderr_line_callback(self, line):
        '''标准错误回调函数
        '''
        self.log.error(line.strip(' '))

    def start(self, retry=1, timeout=CommandTimeout):
        '''启动Agent，类初始化即调用
        
        :param retry: 启动尝试次数
        :type retry: int
        :param timeout: 单次启动超时 (秒)
        :type timeout: int
        '''
        if self.type == EnumDevice.Simulator:
            dt.DT().reboot(self.udid)
            xcode_version = dt.DT.get_xcode_version().split(".")[0]
            if int(xcode_version) >= 9:
                self._agent_cmd = self._get_xcodebuild_agent_cmd_for_simulator(xcode_version)
            else:
                self._agent_cmd = self._get_fbsimctl_agent_cmd_for_simulator(xcode_version)
        else:
            self._agent_cmd = ' '.join(['xcodebuild',
                '-project %s' % self.XCTestAgentPath,
                '-scheme %s' % 'XCTestAgent',
                '-destination "platform=%s,id=%s"' % (self.type, self.udid),
                'test'])
        # 清理遗留的xcodebuild进程
        Process().kill_process_by_name(self._agent_cmd.replace('"', ''))
        Process().kill_process_by_port(self._server_port)
        start_time = time.time()
        # 启动端口转发线程
        self._tcp_relay()
        for _ in range(retry):
            self.log.info("Start XCTestAgent: %s" %self._agent_cmd)
            self.log.info("XCTestAgent Version: %s" %self.version)
            self._process = ThreadTask(command=self._agent_cmd, 
                                       stdout_line_callback=self._stdout_line_callback, 
                                       stderr_line_callback=self._stderr_line_callback)
            
            time0 = time.time()
            while time.time() - time0 < timeout:
                if self.is_working():
                    exec_time = (time.time() - start_time) * 1000
                    self.log.info('[ %s ] consumed [ %dms ]' %('Start XCTestAgent', exec_time))
                    return
                time.sleep(3)
            
            _dt = dt.DT()
            self.log.info("Uninstall com.apple.test.XCTestAgent-Runner")
            result = _dt.uninstall("com.apple.test.XCTestAgent-Runner", self.udid)
            if not result:
                self.log.error(_dt.uninstall_error)
        
        self.stop()
        error_log_name = 'xctest_%s' % self.udid
        agent_error_log = logger.get_agent_error_log(error_log_name, start_time)
        raise AgentStartError("Failed to start XCTestAgent.\nDetails:%s" % agent_error_log)
    
    def stop(self, is_timeout=False):
        '''关闭Agent
        '''
        if not is_timeout:
            # 停止手机上Agent进程
            try:
                self._command_executor.set_timeout(2)
                self.execute(Command.QTA_STOP_AGENT)
            except:
                pass
        # 停止PC上端口转发线程
        if self._relay_thread:
            self._quit_relay_thread = True
            try:
                self._execute(Command.HEALTH) # 发送请求，退出select阻塞
            except:
                pass
            self._is_relay_quit.wait() # 等待端口转发线程退出
            self._relay_thread = None
        self._command_executor.reset_timeout()
        # 等待PC上xcodebuild进程停止
        time.sleep(10)
        self._process = None
        self.session_id = None
        self.log.info("Stop XCTestAgent successfully")
    
    def is_working(self):
        '''检查当前Agent是否在工作，如果Agent无法退到后台，重启手机
        
        :returns: boolean
        '''
        try:
            self._command_executor.set_timeout(1)
            response = self._execute(Command.HEALTH)
            self._command_executor.reset_timeout()
            return response['value'] == 'XCTestAgent is ready'
        except XCTestAgentDeadException:
            self.log.exception('XCTestAgentDead')
            # Agent无法退到后台，重启手机
            try:
                self.log.info('Reboot device %s' % self.udid)
                dt.DT().reboot(self.udid)
                for _ in range(20):
                    device = dt.DT().get_device_by_udid(self.udid)
                    if device is not None:
                        break
                else:
                    raise RuntimeError("after reboot, device disconnected!")
            except Exception:
                self.log.exception('device reboot error')
        
        except XCTestAgentTimeoutException:
            if self._relay_error:
                self.log.error("TcpRelay error:%s" % self._relay_error)
                self.stop()
    
        return False
    
    def has_session(self):
        '''检查是否存在会话
        
        :returns: boolean
        '''
        return self.session_id is not None
    
    def start_session(self, desired_capabilities, timeout=CommandTimeout):
        '''创建新会话
        
        :param desired_capabilities: 
         - bundleId - APP的bundle_id，例如：com.tencent.qq.dailybuild.test
         - app - APP路径，用于运行前安装APP.
        :type desired_capabilities: dict
        :param timeout: 检测启动app超时时间，启动超时则重启agent，单位：秒
        :type timeout: int
        :returns: 返回结果，dict
        '''
        while not self.is_working():
            self.start()
            
        while 1:
            time0 = time.time()
            response = self._execute(Command.SESSION, {
                'desiredCapabilities': desired_capabilities,
            })
            exec_time = int((time.time() - time0) * 1000) # 单位：毫秒
            if exec_time < (timeout*1000):
                break
            self.log.error('Restart XCTestAgent: [ %s ] consumed [ %dms ]' %('start app', exec_time))
            self.stop()
            self.start()
            
        self.session_id = response['sessionId']
        self.capabilities = response['value']['capabilities']
        return response
    
    def _execute(self, command, params={}):
        '''执行指令
        '''
        response = self._command_executor.execute(command, params)
        
        if response:
            if not command in self.LOG_FILTERED_METHODS:
                self.log.debug(response)
            # 检查返回结果
            try:
                self.error_handler.check_response(response)
            except ApplicationCrashedException:
                self.crash_flag = True
                raise
            return response
        # 如果服务器没有发送响应，认为指令执行成功
        return {'success': 0, 'value': None, 'sessionId': self.session_id}
    
    def capture_screen(self):
        try:
            return self.stub_client.execute(Command.SCREENSHOT, {})
        except XCTestAgentTimeoutException:
            return None    
    
    def execute(self, command, params={}, timeout=CommandTimeout):
        '''执行指令

        :param command: 指令名称
        :type command: str
        :param params: 指令参数集
        :type params: dict 例如 {bundleId: com.tencent.test}
        :param timeout: 指令执行超时等待时常
        :type timeout: int
        :returns: 返回结果，dict
        '''
        time0 = time.time()
        
        if not self.has_session():
            response = self._execute(Command.STATUS, {})
            self.session_id = response['sessionId']
        
        if self.has_session():
            if not params:
                params = {'sessionId': self.session_id}
            elif 'sessionId' not in params:
                params['sessionId'] = self.session_id
                
        if command == Command.START:
            if params is None:
                raise WebDriverException("Desired Capabilities can't be None")
            if not isinstance(params, dict):
                raise WebDriverException("Desired Capabilities must be a dictionary")
            response = self.start_session(params, timeout)
            
        elif command == Command.QUIT:
            if 'sessionId' in params:
                response = self._execute(command, params)
            else:
                response = {}
            self.session_id = None
            
        elif command == Command.STATUS:
            response = self._execute(command, params)
            if self.session_id is None:
                self.session_id = response['sessionId']
        else:
            response = self._execute(command, params)
            
        exec_time = int((time.time() - time0) * 1000)
        self.log.info('[ %s ] consumed [ %dms ]' %(command, exec_time))
        return response
        
    def get_crash_flag(self):
        '''获取app是否crash的标识
        
        :returns : boolean -- True if app crashed, False if not
        '''
        return self.crash_flag

    def get_driver_log(self):
        '''获取driver日志路径
        '''
        driver_log_path = None
        for h in self.log.handlers:
            attr_name = 'baseFilename'
            if hasattr(h, attr_name):
                driver_log_path = getattr(h, attr_name)
                break
        return driver_log_path



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''QT4i命令
'''

from __future__ import absolute_import, print_function

import argparse
import os
import logging
import shutil
import sys
import zipfile

from testbase.management import Command
from testbase.management import ArgumentParser
from qt4i.device import DeviceServer
from qt4i.util import less_to
from qt4i.driver.util import zip_decompress
from qt4i.version import version as qt4i_version

DEFAULT_IP = '0.0.0.0'
DEFAULT_PORT = 12306
DEFAULT_AGENT_PORT = 8100
DEFAULT_PID_FILE = '/tmp/driverserver.pid'


class StartDriver(Command):
    '''启动Driver
    '''
    name = 'restartdriver'
    parser = argparse.ArgumentParser("restart local QT4i driver daemon")
    parser.add_argument('--host', '-H', metavar='HOST', dest='host', default=DEFAULT_IP,
                         help='listening address of driver server, local address<%s> is used by default' % DEFAULT_IP)
    parser.add_argument('--port', '-p', metavar='PORT', dest='port', default=DEFAULT_PORT,
                         help='listening port of driver server, port %s is used by default' % DEFAULT_PORT)
    parser.add_argument('--type', '-t', metavar='TYPE', dest='driver_type', default="xctest",
                         help='driver type of driver server(instruments or xctest), "xctest" is used by default')
    parser.add_argument('--pidfile', '-f', metavar='PIDFILE', dest='pidfile', default=DEFAULT_PID_FILE,
                         help='the path of pid file, %s is used by default' % DEFAULT_PID_FILE)
    parser.add_argument('--udid', '-u', metavar='UDID', dest='udid', default=None,
                         help='udid of device')
    parser.add_argument('--agent_port', '-a', metavar='AGENTPORT', dest='agent_port', default=DEFAULT_AGENT_PORT,
                         help='listening port of xctest agent, port %s is used by default' % DEFAULT_AGENT_PORT)
    parser.add_argument('--register', '-r', dest='endpoint_clss', default='',
                         help='the endpoint classes registered to driverserver, for example, "register.demo.Demo,perflib.ios.fps.FPSMonitor"')

    def execute(self, args):
        '''执行过程
        '''
        DeviceServer(args.host, int(args.port), args.udid, int(args.agent_port), args.driver_type, args.endpoint_clss).restart()


class Installer(Command):
    '''安装命令
    '''
    name = 'install'
    parser = argparse.ArgumentParser("install app")
    parser.add_argument("product_path", help="product file path")
    parser.add_argument('-u', default=None, help="device udid", dest="device_udid")

    def execute(self, args):
        '''执行过程
        '''
        from qt4i.driver.tools.dt import DT
        DT().install(args.product_path, args.device_udid)


class Uninstaller(Command):
    '''卸载命令
    '''
    name = 'uninstall'
    parser = argparse.ArgumentParser("uninstall app")
    parser.add_argument("bundle_id", help="app bundle id")
    parser.add_argument('-u', default=None, help="device udid", dest="device_udid")

    def execute(self, args):
        '''执行过程
        '''
        from qt4i.driver.tools.dt import DT
        DT().uninstall(args.bundle_id, args.device_udid)


class DeviceList(Command):
    '''获取当前Mac机上的所有可用设备列表(含真机与模拟器，并产生缓存)
    '''
    name = 'listdevices'
    parser = argparse.ArgumentParser("list devices")

    def execute(self, args):
        '''执行过程
        '''
        from qt4i.driver.tools.dt import DT
        import pprint
        
        for device in DT().get_devices():
            pprint.pprint(device)

            
class Setup(Command):
    '''安装XCTestAgent
    '''
    name = 'setup'
    parser = argparse.ArgumentParser("setup XCTest Agent")

    def execute(self, args):
        '''执行过程
        '''
        egg_file = os.path.join(os.path.dirname(os.path.dirname(__file__)))
        agent_src_path = ""
        if zipfile.is_zipfile(egg_file):
            zip_decompress(egg_file, '/tmp')
            agent_src_path = '/tmp/qt4i/driver/xctest/xctestproj/XCTestAgent'
        else:
            qt4i_path = os.path.dirname(__file__)
            agent_src_path = os.path.join(qt4i_path, 'driver/xctest/xctestproj/XCTestAgent')

        proj_dir = os.path.join(os.path.expanduser('~'), 'XCTestAgent')
        if os.path.exists(proj_dir):
            shutil.rmtree(proj_dir)
        shutil.copytree(agent_src_path, proj_dir)

        proj_path = os.path.join(proj_dir, 'XCTestAgent.xcodeproj')
        os.popen('open %s' % proj_path)

        with open(os.path.join(proj_dir, 'version.txt'), 'w') as f:
            f.write(qt4i_version)

        if os.path.exists('/tmp/qt4i'):
            shutil.rmtree('/tmp/qt4i')


class Update(Command):
    '''更新XCTestAgent
    '''
    name = 'update'
    parser = argparse.ArgumentParser("update XCTest Agent")

    def execute(self, args):
        '''执行过程
        '''
        egg_file = os.path.join(os.path.dirname(os.path.dirname(__file__)))
        proj_dir = os.path.join(os.path.expanduser('~'), 'XCTestAgent')
        is_updated = False
        if zipfile.is_zipfile(egg_file):
            zip_decompress(egg_file, '/tmp')
            agent_src_path = '/tmp/qt4i/driver/xctest/xctestproj/XCTestAgent'
        else:
            qt4i_path = os.path.dirname(__file__)
            agent_src_path = os.path.join(qt4i_path, 'driver/xctest/xctestproj/XCTestAgent')
        if os.path.exists(proj_dir):
            with open(os.path.join(proj_dir, 'version.txt'), 'r') as f:
                version = f.read()
            if not less_to(version, qt4i_version):
                return
            else:
                lib_dir = os.path.join(proj_dir, 'third')
                shutil.rmtree(lib_dir)
                lib_src_path = os.path.join(agent_src_path, 'third')
                shutil.copytree(lib_src_path, lib_dir)
                is_updated = True
        else:
            shutil.copytree(agent_src_path, proj_dir)
            proj_path = os.path.join(proj_dir, 'XCTestAgent.xcodeproj')
            os.popen('open %s' % proj_path)
            is_updated = True

        if is_updated:
            with open(os.path.join(proj_dir, 'version.txt'), 'w') as f:
                f.write(qt4i_version)

            if os.path.exists('/tmp/qt4i'):
                shutil.rmtree('/tmp/qt4i')


class RebootDevice(Command):
    '''重启手机
    '''
    name = 'reboot'
    parser = argparse.ArgumentParser("reboot device")
    parser.add_argument('-u', default=None, help="device udid", dest="device_udid")

    def execute(self, args):
        '''执行过程
        '''
        from qt4i.driver.tools.dt import DT
        DT().reboot(args.device_udid)


class ListAppsOnDevice(Command):
    '''获取手机上的app列表
    '''

    name = 'listapps'
    parser = argparse.ArgumentParser("List applications of the specified device")
    parser.add_argument('-u', default=None, help="device udid", dest="device_udid")
    parser.add_argument('-t', default='user', help="app type(user/system/all)", dest="app_type")

    def execute(self, args):
        '''执行过程
        '''
        from qt4i.driver.tools.dt import DT
        import pprint
        pprint.pprint(DT().list_apps(args.device_udid, args.app_type))


class UpdateFbsimctl(Command):
    '''更新fbsimctl
    '''

    name = 'updatefbsimctl'
    parser = argparse.ArgumentParser("Update fbsimctl tool in /cores/fbsimctl")

    def execute(self, args):
        '''执行过程
        '''
        fbsimctl_path = '/cores/fbsimctl'
        if os.path.exists(fbsimctl_path):
            shutil.rmtree(fbsimctl_path)


def qt4i_manage_main():
    logging.root.addHandler(logging.StreamHandler())
    logging.root.level = logging.INFO
    cmds = [StartDriver, Installer, Uninstaller, DeviceList, Setup, Update, \
        RebootDevice, ListAppsOnDevice, UpdateFbsimctl]
    argparser = ArgumentParser(cmds)
    subcmd, args = argparser.parse_args(sys.argv[1:])
    subcmd.execute(args)



# -*- coding:utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''资源的管理和分配调度工具
'''

from __future__ import absolute_import, print_function

import threading


MAX_PORT_NUM = 50

class PortManager(object):
    '''端口管理器基类
    '''
    _port_maps = {}  # 维护映射端口字典的集合，key为'agent'或'web'
    _lock = threading.Lock()
    EnumPortType = ['agent', 'web']

    @classmethod
    def set_port(cls, port_type, udid, port=None, base=8100):
        '''设置设备udid对应的端口
        
        :param port_type: 端口类型
        :type port_type: EnumPortType
        :param udid: 设备udid
        :type udid: str
        :param port: 端口号
        :type port: int
        :param base: 基准端口
        :type base: int
        '''
        if port_type not in PortManager.EnumPortType:
            raise Exception("端口类型%s异常，正确的类型有：%s" % (port_type,  str(cls.EnumPortType)))
        if port_type not in cls._port_maps:
            with cls._lock:
                cls._port_maps[port_type] = {}
        if not port:
            port = base
        if udid not in cls._port_maps[port_type]:
            hash_port = hash(udid) % MAX_PORT_NUM
            while 1:
                with cls._lock:
                    if port not in cls._port_maps[port_type].values():
                        cls._port_maps[port_type][udid] = port
                        break
                port = base + hash_port
                hash_port = (hash_port + 1) % MAX_PORT_NUM
    
    @classmethod
    def get_port(cls, port_type, udid):
        '''通过设备udid获取对应的端口
        
        :param port_type: 端口类型
        :type port_type: EnumPortType
        :param udid: 设备udid
        :type udid: str
        '''
        if port_type not in cls.EnumPortType:
            raise Exception("端口类型%s异常，正确的类型有：%s" % (port_type,  str(cls.EnumPortType)))
        if port_type not in cls._port_maps:
            with cls._lock:
                cls._port_maps[port_type] = {}
        if udid not in cls._port_maps[port_type]:
            if port_type == 'web':
                cls.set_port(port_type, udid, base=27753)
            else:
                cls.set_port(port_type, udid)
        return cls._port_maps[port_type][udid]
    
    @classmethod
    def exist(cls, port_type, udid):
        '''判断端口字典中是否已存在该设备
        
        :param port_type: 端口类型
        :type port_type: EnumPortType
        :param udid: 设备udid
        :type udid: str
        '''
        if port_type not in cls.EnumPortType:
            raise Exception("端口类型%s异常，正确的类型有：%s" % (port_type,  str(cls.EnumPortType)))
        if port_type not in cls._port_maps:
            with cls._lock:
                cls._port_maps[port_type] = {}
        return udid in cls._port_maps[port_type]
    
    @classmethod
    def del_port(cls, port_type, udid):
        '''从字典中删除设备及端口
        
        :param port_type: 端口类型
        :type port_type: EnumPortType
        :param udid: 设备udid
        :type udid: str
        '''
        if port_type not in cls.EnumPortType:
            raise Exception("端口类型%s异常，正确的类型有：%s" % (port_type,  str(cls.EnumPortType)))
        if port_type not in cls._port_maps:
            with cls._lock:
                cls._port_maps[port_type] = {}
        if udid in cls._port_maps[port_type]:
            with cls._lock:
                cls._port_maps[port_type].pop(udid)
    
    @classmethod
    def ports(cls, port_type):
        '''返回端口字典
        
        :param port_type: 端口类型
        :type port_type: EnumPortType
        '''
        if port_type not in cls.EnumPortType:
            raise Exception("端口类型%s异常，正确的类型有：%s" % (port_type,  str(cls.EnumPortType)))
        if port_type not in cls._port_maps:
            with cls._lock:
                cls._port_maps[port_type] = {}
        return cls._port_maps[port_type]


# coding: UTF-8
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''此模块用于归纳Xcode
'''
# -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*-
from __future__ import absolute_import, print_function

import os, re
from qt4i.driver.util import FileManager
from qt4i.driver.util import Task
from qt4i.driver.util import ThreadTask
CurrentDirectoryPath = os.path.abspath(os.path.dirname(__file__))
# -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*- -*-

class Xcode(object):
    
    DeveloperPath = Task("xcode-select -p").execute()
    InstrumentsPath = Task("xcrun -f instruments").execute()
    InstrumentsTracetemplatePath = os.path.join(CurrentDirectoryPath, '_tracetemplate')
    InstrumentsTraceTemplatesPath = {'xcode5_default' : os.path.abspath('%s/../Applications/Instruments.app/Contents/PlugIns/AutomationInstrument.bundle/Contents/Resources/Automation.tracetemplate' % DeveloperPath),
                                     'xcode6_default' : os.path.abspath('%s/../Applications/Instruments.app/Contents/PlugIns/AutomationInstrument.xrplugin/Contents/Resources/Automation.tracetemplate' % DeveloperPath)}

    def __init__(self):
        self.instruments_command = None

    def __get_instruments_path__(self, simulator=False):
        return self.InstrumentsPath

    def __get_instruments_trace_templates_path__(self):
        _items = FileManager(self.InstrumentsTracetemplatePath).get_children()
        for _item in _items:
            _cur_name, _cur_ext = os.path.splitext(os.path.basename(_item))
            if re.match('^.tracetemplate$', _cur_ext, re.I):
                self.InstrumentsTraceTemplatesPath.update({_cur_name:_item})
        return self.InstrumentsTraceTemplatesPath
    
    def __get_instruments_trace_template__(self, _name_or_abspath=None):
        if _name_or_abspath and os.path.exists(_name_or_abspath):
            return _name_or_abspath
        _trace_templates_path = self.__get_instruments_trace_templates_path__()
        _trace_template = _trace_templates_path.get(_name_or_abspath)
        if _trace_template and os.path.exists(_trace_template):
            return _trace_template
        else:
            for _name in ['xcode6_default', 'xcode5_default']:
                _trace_template = _trace_templates_path.get(_name)
                if os.path.exists(_trace_template):
                    return _trace_template
        if None == _trace_template or False == os.path.exists(_trace_template):
            raise Exception('tracetemplate is invalid')
    
    def start_instruments(self, device_udid, device_simulator, bundle_id, uia_script, uia_results_path, trace_template=None, started_callback=None, stdout_line_callback=None, stderr_line_callback=None, return_code_callback=None, cwd=None, env=None, print_enable=False, **args):
        if not os.path.exists(self.InstrumentsPath):
            raise Exception('instruments is invalid.')
        if device_udid is None:
            raise Exception('device_udid is invalid.')
        if bundle_id is None:
            raise Exception('bundle_id is invalid.')
        if uia_script is None or os.path.exists(uia_script) == False:
            raise Exception('uia_script is invalid.')
        if uia_results_path is None or os.path.exists(uia_results_path) == False:
            raise Exception('uia_results_path is invalid.')
        instruments = self.__get_instruments_path__(device_simulator)
        trace_template = self.__get_instruments_trace_template__(trace_template)
        cwd = cwd or CurrentDirectoryPath
                 
        self.instruments_command = " ".join([
                            '%s' % instruments,
                            '-w "%s"' % device_udid,
                            '-t "%s"' % trace_template,
                            '' if args['trace_output'] is None else '-D "%s"' % args['trace_output'],
                            '"%s"' % bundle_id,
                            '-e UIASCRIPT "%s"' % uia_script,
                            '-e UIARESULTSPATH "%s"' % uia_results_path,
                            '-v',
                            args['params'] if 'params' in args else ''
                           ])
        return ThreadTask(self.instruments_command, started_callback, stdout_line_callback, stderr_line_callback, return_code_callback, cwd, env, print_enable)



# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''此模块用于归纳基础通用子模块
'''

import os, re, stat, shutil
from past.builtins import xrange


class FileManager(object):
    '''文件相关操作类
    '''
    
    def __init__(self, path):
        self.path = os.path.abspath(path)
    
    def exists(self):
        return os.path.exists(self.path)
    
    def repair_path(self, path=None):
        path = os.path.abspath(path or self.path)
        if not os.path.exists(path): os.makedirs(path)
        return os.path.exists(path)
    
    def __sort_by_ctime__(self, items):
        length = len(items)
        for i in xrange(length - 1, 0, -1):
            flag = 0
            for j in xrange(0, i):
                ctime1 = os.path.getctime(items[j])
                ctime2 = os.path.getctime(items[j + 1])
                if ctime1 > ctime2:
                    items[j], items[j + 1] = items[j + 1], items[j]
                    flag += 1
            if flag == 0 : break;
        return items
    
    def get_children(self, path=None, recursion=False, sort_by_ctime=False, name_filter=['.DS_Store']):
        items, path = [], os.path.abspath(path or self.path)
        if recursion:
            for dirpath, dirnames, filenames in os.walk(path) : items += [os.path.join(dirpath, name) for name in dirnames + filenames if name not in name_filter and not re.match('|'.join(name_filter), name, re.IGNORECASE)]
        else  : items = [os.path.join(path, item) for item in os.listdir(path) if item not in name_filter and not re.match('|'.join(name_filter), item, re.IGNORECASE)]
        if sort_by_ctime : items = self.__sort_by_ctime__(items)
        return items
    
    def force_delete(self, path=None):
        path = os.path.abspath(path or self.path)
        if path is None or os.path.exists(path) == False : return
        if os.path.isfile(path):
            os.chmod(path, stat.S_IWRITE)
            os.remove(path)
            return
        try    : shutil.rmtree(path)
        except :
            del_dir_paths = []
            for dirpath, dirnames, filenames in os.walk(path): #  #@UnusedVariable
                try    : shutil.rmtree(dirpath)
                except :
                    del_dir_paths.append(dirpath)
                    for filename in filenames:
                        del_file_path = os.path.join(dirpath, filename)
                        os.chmod(del_file_path, stat.S_IWRITE)
                        os.remove(del_file_path)
            del_dir_paths.sort(reverse=True)
            for del_dir_path in del_dir_paths:
                os.chmod(del_dir_path, stat.S_IWRITE)
                os.rmdir(del_dir_path)
    
    def force_copy_to(self, dst):
        # if os.path.exists(dst) and os.path.isdir(dst):
        #    dst = os.path.join(dst, os.path.basename(self.path))
        if os.path.exists(dst):
            self.force_delete(dst)
        self.repair_path(os.path.dirname(dst))
        if os.path.isfile(self.path):
            shutil.copy(self.path, dst)
        if os.path.isdir(self.path):
            shutil.copytree(self.path, dst)

def zip_decompress(zip_file, output_dir):
    '''解压zip文件到指定目录
    
    :param zip_file: 需解药的zip文件路径
    :param output_dir: 解压后保存的目录路径
    '''
    import zipfile
    if not zipfile.is_zipfile(zip_file):
        raise Exception('not support to decompress non-zipfile')
    
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, 0o777)
        
    f_zip = zipfile.ZipFile(zip_file, 'r')
    f_zip.extractall(output_dir)
    f_zip.close()